#include <centralrepository.h>
#include <EqualizerConstants.h>
#include <AudioEqualizerUtility.h>

#include "EffectEngine.h"

CEqPresetNotifyHandler* CEqPresetNotifyHandler::NewL(MEqPresetWatcher& aObserver)
{
	CEqPresetNotifyHandler* self = new (ELeave) CEqPresetNotifyHandler(aObserver);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);
	return self;
}

void CEqPresetNotifyHandler::ConstructL()
{
        iUid = KCRUidMPXMPSettings; //observe Preset engine to
        iKey = KMPXMPEqPresetId;    //see which Preset is active
        iRepository = CRepository::NewL( iUid );
        CActiveScheduler::Add(this);
}

CEqPresetNotifyHandler::CEqPresetNotifyHandler(MEqPresetWatcher& aObserver)
 :CActive(CActive::EPriorityStandard),
  iObserver(aObserver)
{
         //do nothing
}

CEqPresetNotifyHandler::~CEqPresetNotifyHandler()
{
         Cancel(); //first cancel because iRepository is used there
         delete iRepository;
         iRepository=NULL;
}

void CEqPresetNotifyHandler::StartObservingL()
{
         if( IsActive() )
         {
         return; //do nothing if allready observing
         }
         User::LeaveIfError( iRepository->NotifyRequest( iKey, iStatus ) );
         SetActive();
}

void CEqPresetNotifyHandler::StopObserving()
{
          Cancel();
}

void CEqPresetNotifyHandler::DoCancel()
{
          iRepository->NotifyCancel(iKey);
}

void CEqPresetNotifyHandler::RunL()
{
          TInt presetId( KEqualizerPresetNone );

          TInt error = iRepository->Get( iKey, presetId);

          //always check the error code
          if( error == KErrNone )
                 {
                 //observer received the Preset value as integer
                 //this happens even if the value hasn't really changed
                 //some application may have set the same value again
                 iObserver.EqPresetChanged(presetId);
                 }

          // Re-subscribe
          error = iRepository->NotifyRequest( iKey, iStatus );
          if( error == KErrNone )
               {
               //
               SetActive();
               }
          else
               {
                //todo: error handling
               }
}

///////////////////////////////////////////////////////////////////////

CAudioEffectEngine* CAudioEffectEngine::NewL( CMdaAudioPlayerUtility* aPlayer, TBool aDolbySupported, MMyDolbyHeadphoneObserver2& aObs) 
{
    CAudioEffectEngine* self = new( ELeave ) CAudioEffectEngine(aDolbySupported, aObs);
    CleanupStack::PushL( self );
    self->ConstructL(aPlayer);
    CleanupStack::Pop( self );
    return self;
}


CAudioEffectEngine* CAudioEffectEngine::NewL( CMdaAudioOutputStream* aAudioOutput, TBool aDolbySupported, MMyDolbyHeadphoneObserver2& aObs)
{
    CAudioEffectEngine* self = new( ELeave ) CAudioEffectEngine(aDolbySupported, aObs);
    CleanupStack::PushL( self );
    self->ConstructL(aAudioOutput);
    CleanupStack::Pop( self );
    return self;
}

 
CAudioEffectEngine::CAudioEffectEngine(TBool aDolbySupported, MMyDolbyHeadphoneObserver2& aObs)
 : iDolbySupported(aDolbySupported),
   iObs(aObs)
{
    iMdaPlayer   = NULL; // Not owned
    iAudioOutput = NULL; // Not owned
    iType = EFFPtNone;

    iDolbyHeadphone = NULL;
}


CAudioEffectEngine::~CAudioEffectEngine()
{    
    DestroyAudioEffect();

    if(iDolbyHeadphone) delete iDolbyHeadphone;
}


void CAudioEffectEngine::DolbyConstructL()
{
    if(iDolbySupported)
    {
      iDolbyHeadphone = CMyDolbyHeadphone::NewL(*this);
    }
}

void CAudioEffectEngine::ConstructL(CMdaAudioPlayerUtility* aPlayer)
{
    iMdaPlayer = aPlayer;
    iType = EFFPtExternal;

    DolbyConstructL();
}

void CAudioEffectEngine::ConstructL(CMdaAudioOutputStream* aAudioOutput)
{
    iAudioOutput = aAudioOutput;
    iType = EFFPtInternal;

    DolbyConstructL();
}

void CAudioEffectEngine::SetEqualizerL(TInt aPreset)
{
    // Create the effect
    //
    if( !iEqualizerEffect )
     {
	switch(iType)
	{
	case EFFPtExternal:
          iEqualizerEffect = CAudioEqualizerUtility::NewL(*iMdaPlayer);
	break;

	case EFFPtInternal:
          iEqualizerEffect = CAudioEqualizerUtility::NewL(*iAudioOutput);
	break;

	default: return;
	}
     }
        
    // Get equalizer preset id from cenrep
    //
    TInt presetId( aPreset );

    if(presetId == KEqualizerPresetNone)
    {
      TRAP_IGNORE(
        {
        CRepository* repository = CRepository::NewL( KCRUidMPXMPSettings );
        repository->Get( KMPXMPEqPresetId, presetId );
        delete repository;
        repository = NULL;
        } );
    }

    // Translate preset ID to index
    //
    TInt index( KErrNotFound );
    if ( iEqualizerEffect && presetId != KEqualizerPresetNone )
        {
        TArray<TEfAudioEqualizerUtilityPreset> presetArray = iEqualizerEffect->Presets();

        TBool found( EFalse );
        for ( TInt i = 0; i < presetArray.Count() && found == EFalse; i++ )
            {
            if ( presetArray[i].iPresetNameKey == 
                static_cast<TUint32>( presetId ) )
                {
                found = ETrue;
                index = i;
                }
            }
        }
    
    // Apply preset index index or disable if not found
    //
    if ( KErrNotFound != index )
        {
        iEqualizerEffect->ApplyPresetL( index );
        }
    else
        {
        iEqualizerEffect->DisableEqualizerL();    
        }
}

void CAudioEffectEngine::SetDolbyHeadphoneL(TBool aOn)
{
    TInt err(KErrNone);

    if(iDolbyHeadphone && !iDolbyHeadphone->IsCreated())
    {
	switch(iType)
	{
	case EFFPtExternal:
          TRAP(err, iDolbyHeadphone->CreateDolby1L(*iMdaPlayer));
	break;

	case EFFPtInternal:
          TRAP(err, iDolbyHeadphone->CreateDolby2L(*iAudioOutput));
	break;

	default: err = -1;
	}
    }

    if(err == KErrNone && iDolbyHeadphone->IsCreated())
      iDolbyHeadphone->DolbyOnL(aOn);
}
    
void CAudioEffectEngine::CreateAudioEffectsL()
{
    SetEqualizerL(KEqualizerPresetNone);
}
    
void CAudioEffectEngine::DestroyAudioEffect()
{
    if (iEqualizerEffect)
    {
        delete iEqualizerEffect;  //effect automatically disabled
        iEqualizerEffect = NULL;
    }

    if(iDolbyHeadphone && iDolbyHeadphone->IsCreated())
    {
       iDolbyHeadphone->DestroyDolby();

       iObs.DolbyChange2(EFalse);
    }
}

void CAudioEffectEngine::DolbyChange(TBool aState)
{
   iObs.DolbyChange2(aState);
}
