// INCLUDE FILES
#include <f32file.h>
#include <s32file.h>
#include <bautils.h>
#include <es_sock.h>
#include <e32math.h>			// for Pow

#include <remconcoreapitarget.h>

#include "FolderPlayView.h"
#include "FolderPlayEngine.h"


const TInt KBufferSize = 0x10000; //64k - ~1sec for 44,1kHz

// mp2 audio FourCC, if supported
const TUint32 KMMFFourCCCodeMP2 = 0x32504d20;	//(' ','M','P','2')

	
CPlayEngine* CPlayEngine::NewL(CFolderView& aParent, RFs& aFs)
{
	CPlayEngine* self = new (ELeave) CPlayEngine(aParent, aFs);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);
	return self;
}


void CPlayEngine::ConstructL()
{
	iAudioOutput = CMdaAudioOutputStream::NewL(*this);
	iAudioOutput->SetPriority(EMdaPriorityMax, EMdaPriorityPreferenceQuality); //EMdaPriorityNormal

	iAudioEffects = CAudioEffectEngine::NewL( iAudioOutput, iParent.IsDolbySupported(), *this);

	iEvenBuffer.CreateMaxL(KBufferSize);
	iOddBuffer.CreateMaxL(KBufferSize);

	iEngineBuffers.CreateMaxL(sizeof(Engine_buffers));
	
	CActiveScheduler::Add(this);
}


CPlayEngine::CPlayEngine(CFolderView& aParent, RFs& aFs)
 :CActive(CActive::EPriorityHigh),
 iFs(aFs),
 iParent(aParent),
 iCurBuffer(NULL, 0)
{
	iAudioBusy = EFalse;
	
	iMainThread.Open(RThread().Id());
	iMainThread.SetPriority(EPriorityAbsoluteForeground);
	
	iAudioOutput = NULL;
	
	iExclusive.CreateLocal();
	iStateChange.CreateLocal();
	iFreeBuffer.CreateLocal(1);
}


CPlayEngine::~CPlayEngine()
{
	Stop();
	
	Cancel();
	
        delete iAudioEffects;

	if(iAudioOutput) { delete iAudioOutput; iAudioOutput = NULL; }

	iEvenBuffer.Close();
	iOddBuffer.Close();

	iExclusive.Close();
	iStateChange.Close();
	iFreeBuffer.Close();

	iEngineBuffers.Close();
	
	iMainThread.Close();
}


////////////////////////////////////////////////////////////////////////////////
// folded FourCC (big-endian -> little-endian)

static inline TFourCC operator ~(TFourCC anFourCC)
{
	return anFourCC == KMMFFourCCCodePCM16B ? KMMFFourCCCodePCM16 :
		(anFourCC == KMMFFourCCCodePCMU16B ? KMMFFourCCCodePCMU16 : anFourCC);
}


TInt CPlayEngine::read_filebuf(TUint8* pBuffer, int nBytesToRead)
{
    TPtr8 aBuffer(pBuffer, nBytesToRead);
    if(iAudioFile.Read(aBuffer, nBytesToRead) != KErrNone)
 	return -1;
    else
	return aBuffer.Length();
}


////////////////////////////////////////////////////////////////////////////////
// outputs the current buffer to the audio device,
// waits for the next available buffer and switchs to it
//
// virtually blocks

void CPlayEngine::OutputBuffer()
{
	iExclusive.Wait();
	// freeze, if playback is paused
	if(iParent.iPause)
	{
		TBool aSetAudioParam = iSetAudioParam;
		if(!aSetAudioParam) iCurBuffer.Set(InactiveBuffer().MidTPtr(0));	// if possible, switch back to the previous buffer

		while(iParent.iPause && iParent.iAlive) iStateChange.Wait(iExclusive);						// wait here for resume
		if(!iParent.iAlive) { iExclusive.Signal(); return; }
		if(!aSetAudioParam)																								// if switched to the previous buffer, send it, and switch back
		{
			// send the last returned buffer once again
			iStatus = KRequestPending; SetActive(); iBufferPending = ETrue;
			TRequestStatus* aStatus = &iStatus; iMainThread.RequestComplete(aStatus, KErrNone);
			while(iBufferPending && iParent.iAlive) iStateChange.Wait(iExclusive);
			// dereference buffer resource and switch to another buffer
			iFreeBuffer.Wait(); iCurBuffer.Set(InactiveBuffer().MidTPtr(0));
		}
	}
	
// if FourCC is scheduled to be changed, wait for audio device to empty out
	if(iSetAudioParam)
	{
		while(iAudioBusy && iParent.iAlive) if(iStateChange.TimedWait(iExclusive, 1000000) == KErrTimedOut) iAudioBusy = EFalse;
		if(!iParent.iAlive) { iExclusive.Signal(); return; }
	}

// boost PCM audio signal on-the-fly

//	if(iParent.iIsBoostVol && iAudioDataType == KMMFFourCCCodePCM16)
	if(iParent.iBoostVolCur && iAudioDataType == KMMFFourCCCodePCM16)
	{
		TInt16* aSample = (TInt16*)&ActiveBuffer()[0]; TInt16* aEnd = aSample + (ActiveBuffer().Length() >> 1);
		TReal aFactor; Math::Pow(aFactor, 10, (TReal)iParent.iBoostVal / 10);
		while(aSample < aEnd)
		{
			TInt aBoosted = (TInt)(*aSample * aFactor + (*aSample >= 0 ? .5 : -.5));
			*aSample = aBoosted >= 0x8000 ? 0x7FFF : (aBoosted < -0x8000 ? 0x8000 : (TInt16)aBoosted);
			++aSample;
		}
	}

// write to audio device in the main thread (RunL)
	iStatus = KRequestPending; SetActive(); iBufferPending = ETrue;
	TRequestStatus* aStatus = &iStatus; iMainThread.RequestComplete(aStatus, KErrNone);
	while(iBufferPending && iParent.iAlive) iStateChange.Wait(iExclusive);
		
	iExclusive.Signal();
// wait for an empty buffer and switch to it
	iFreeBuffer.Wait();
	InactiveBuffer().SetMax(); iCurBuffer.Set(InactiveBuffer().MidTPtr(0));
}

////////////////////////////////////////////////////////////////////////////////
// flushes remaining audio data
//
// <iCurBuffer> must point to the remaining portion of a physical buffer
// virtually blocks

void CPlayEngine::Flush()
{
	if(iCurBuffer.MaxLength() < iEvenBuffer.MaxLength())	// if there is some data pending
	{
		ActiveBuffer().SetLength(iEvenBuffer.MaxLength() - iCurBuffer.MaxLength());
		OutputBuffer();
	}
}

///////////////////////////////////////////////////
struct MP3Header
{
	TUint8 sync_hi;			// sync word high (all set)

	TUint8 CRC			: 1;	// 0 - Protected by CRC (16bit CRC follows header)
	TUint8 layer			: 2;	// 00 - reserved, 01 - Layer III, 10 - Layer II, 11 - Layer I
	TUint8 version			: 2;	// 00 - MPEG Version 2.5 (unofficial), 01 - reserved, 10 - MPEG Version 2, 11 - MPEG Version 1
	TUint8 sync_lo			: 3;	// sync word low  (all set)

	TUint8				: 1;	// private
	TUint8 padding			: 1;	// 1 - frame is padded with one extra slot
	//	bits	MPEG1		MPEG2		MPEG2.5 
	//	00	44100		22050		11025 
	//	01	48000		24000		12000 
	//	10	32000		16000		8000 
	//	11					reserved 
	TUint8 sample_rate		: 2;
	TUint8 bitrate			: 4;

	TUint8 emphasis			: 2;	// 00 - none, 01 - 50/15 ms, 10 - reserved, 11 - CCIT J.17
	TUint8 original			: 1;
	TUint8 copyright		: 1;
	TUint8 mode_ext			: 2;	// only used in Joint Stereo
	TUint8 channel_mode		: 2;	// 00 - Stereo, 01 - Joint stereo (Stereo), 10 - Dual channel (2 mono channels), 11 - Single channel (Mono)

	inline TFourCC DataType() const
	{ return layer == 0x1 ? KMMFFourCCCodeMP3 : (layer == 0x2 ? KMMFFourCCCodeMP2 : KMMFFourCCCodeNULL); }
		
	inline TInt SampleRate() const
	{
		static const TInt ac[3][4] =
		{
			{ TMdaAudioDataSettings::ESampleRate11025Hz, 0, TMdaAudioDataSettings::ESampleRate22050Hz, TMdaAudioDataSettings::ESampleRate44100Hz },
			{ TMdaAudioDataSettings::ESampleRate12000Hz, 0, TMdaAudioDataSettings::ESampleRate24000Hz, TMdaAudioDataSettings::ESampleRate48000Hz },
			{ TMdaAudioDataSettings::ESampleRate8000Hz, 0, TMdaAudioDataSettings::ESampleRate16000Hz, TMdaAudioDataSettings::ESampleRate32000Hz }
		};
		
		return sample_rate != 0x3 ? ac[sample_rate][version] : 0;
	}
		
	inline TInt Channels() const
	{ return channel_mode != 0x3 ? TMdaAudioDataSettings::EChannelsStereo : TMdaAudioDataSettings::EChannelsMono; }
};

bool CPlayEngine::ParseMP3()
{
	static TBuf8<1024> aBuffer;
	if(iParent.iCurrentPos < 0) iParent.iCurrentPos = 0;
	
	// if starts with ID3, skip it
	iAudioFile.Read(aBuffer, 10); TInt aPos;
	if (aBuffer.Length() == 10 && aBuffer.Left(3) == _L8("ID3") &&
		(aBuffer[6] & 0x80) == 0 && (aBuffer[7] & 0x80) == 0 &&
		(aBuffer[8] & 0x80) == 0 && (aBuffer[9] & 0x80) == 0)
	{
		aPos = ((TInt)aBuffer[6] << 21) | ((TInt)aBuffer[7] << 14) | ((TInt)aBuffer[8] << 7) | aBuffer[9];
		iAudioFile.Seek(ESeekCurrent, aPos);
	}
	else { aPos = -aBuffer.Length(); iAudioFile.Seek(ESeekCurrent, aPos); }
	
	while(true)
	{
		if(!iParent.iAlive) return true;
		iAudioFile.Read(aBuffer, aBuffer.MaxLength()); if(aBuffer.Length() < 4) return false;
		TInt aSync = aBuffer.Locate(0xFF);
		if(aSync != KErrNotFound)
		{
			MP3Header aAudioParam; TPtr8 aPtr((TUint8*)&aAudioParam, 4); aPos += aSync;
			iAudioFile.Seek(ESeekStart, aPos); iAudioFile.Read(aPtr, 4); if(aPtr.Length() < 4) return false;
			if(aAudioParam.sync_lo == 0x7)
			{
				iDataType = aAudioParam.DataType(); iSampleRate = aAudioParam.SampleRate(); iChannels = aAudioParam.Channels();
				break;
			}
			else iAudioFile.Seek(ESeekStart, aPos += 2);
		}
		else aPos += aBuffer.Length();
	}

	if(iDataType == KMMFFourCCCodeNULL || iSampleRate == 0 || iChannels == 0) return false;
	ScheduleAudioReset();
	
	TInt aSize; iAudioFile.Size(aSize);

	iParent.SetTotalLength(aSize - aPos);

	aPos += iParent.iCurrentPos; iAudioFile.Seek(ESeekStart, aPos);

	TInt aDataLen = aSize - aPos;

	while(iParent.iAlive)
	{
		iAudioFile.Read(iCurBuffer, Min(aDataLen, iCurBuffer.MaxLength()));

//		iParent.iCurrentPos += iCurBuffer.Length();
		
		if(iCurBuffer.Length() < iCurBuffer.MaxLength()) break;
		aDataLen -= iCurBuffer.Length();
		
		OutputBuffer();

		iParent.iCurrentPos += iCurBuffer.Length(); // dsv
	}
	aDataLen = iCurBuffer.Length();

//	iParent.iCurrentPos += iCurBuffer.Length(); // dsv

	iCurBuffer.SetMax();
	iCurBuffer.Set(iCurBuffer.MidTPtr(aDataLen));

	return true;
}


//////////////////////////////////////////////////////////////////////////////////
// if audio parameters have changed, schedules audio reset as soon as possible

void CPlayEngine::ScheduleAudioReset()
{
	if(~iDataType != iAudioDataType || iSampleRate != iAudioSettings.iSampleRate || iChannels != iAudioSettings.iChannels)
	{
		Flush(); iSetAudioParam = ETrue;
		iAudioDataType = ~iDataType;
		iAudioSettings.iSampleRate = iSampleRate; iAudioSettings.iChannels = iChannels;
	}
}


bool CPlayEngine::ParseFile(const TDesC16& anFile)
{
//!!!	iAudioFile.Close();

	TInt aExtension = anFile.LocateReverse(L'.');
	if(aExtension == KErrNotFound) return false;
	
	if(iAudioFile.Open(iFs, anFile, EFileRead|EFileShareReadersOnly) != KErrNone) return false;

// update visual status
	iStatus = KRequestPending; SetActive(); iBufferPending = EFalse;
	TRequestStatus* aStatus = &iStatus; iMainThread.RequestComplete(aStatus, KErrNone);
	
	TPtrC16 aType(anFile.Mid(aExtension + 1));

	if(aType.CompareF(_L16("flac")) == 0) return ParseFLAC();
	else if(aType.CompareF(_L16("wv")) == 0) return ParseWavPack();
	else if(aType.CompareF(_L16("ape")) == 0 || aType.CompareF(_L16("mac")) == 0) return ParseMAC();
	else if(iParent.iQuasiLameMP3 && aType.CompareF(_L16("mp3")) == 0) return ParseMP3();
	else if(aType.CompareF(_L16("ogg")) == 0) return ParseOgg();
	
	return true;
}

void CPlayEngine::PlayTheList()
{
	while(iParent.iCurFileIdx < iParent.iPlayList.Count())
	{
		ParseFile(iParent.iPlayList[iParent.iCurFileIdx]);
		iAudioFile.Close(); //!!!

		if(!iParent.PlayNextInList()) return;

		if((iParent.iIsRepeatly || iParent.iIsRepeatlyOne) &&
		   (iParent.iCurFileIdx == iParent.iPlayList.Count()))
		{
			iParent.iCurFileIdx = 0;
			continue;
		}
	}
	Flush();


// wait till playback finishes	
	iExclusive.Wait();
	while(iAudioBusy && iParent.iAlive) if(iStateChange.TimedWait(iExclusive, 2000000) == KErrTimedOut) iAudioBusy = EFalse;
	iExclusive.Signal();

// clean up
	iAudioThread.Close();

	iAudioFile.Close();

// update visual status
	iStatus = KRequestPending; SetActive();
	iBufferPending = EFalse; iLastFilePlayed = ETrue;
	TRequestStatus* aStatus = &iStatus; iMainThread.RequestComplete(aStatus, KErrNone);
}

static TInt PlayerFunc(TAny* aPlayEngine)
{
	TInt ret(KErrNone);
	CTrapCleanup* cleanup = CTrapCleanup::New();

	TRAP(ret, ((CPlayEngine*)aPlayEngine)->PlayTheList())

	delete cleanup;

	return ret;
}

///////////////////////////////////////////////////////////////////////
// events in the main thread's context

void CPlayEngine::MaoscOpenComplete(TInt aError)
{
	if(aError != KErrNone) return;

	SetVolume(iParent.iCurrentVol);

	TInt err(KErrNone);
	TInt err2(KErrNone);

        TRAP(err, iAudioEffects->CreateAudioEffectsL());

	if (iParent.IsDolbySupported() && iParent.IsFMTxPassive())
	{
	  TRAP(err2, iAudioEffects->SetDolbyHeadphoneL(iParent.GetDolbyState()));
	}


        if( err != KErrNone || err2 != KErrNone)
        {
	  iAudioEffects->DestroyAudioEffect();

          TRAP(err, iAudioEffects->CreateAudioEffectsL());

 	  if (iParent.IsDolbySupported() && iParent.IsFMTxPassive())
	  {
	  TRAP(err2, iAudioEffects->SetDolbyHeadphoneL(iParent.GetDolbyState()));
	  }
        }

	if(iParent.iAlive && !iParent.iPause && iBufferPending) { iStatus = KErrNone; RunL(); return; }
	
	iFreeBuffer.Close(); iFreeBuffer.CreateLocal(1);		// buffer availability sync
	
	if(iParent.iAlive && iParent.iPause) // Resume
	{
		iSetAudioParam = ETrue;
		iExclusive.Wait(); iParent.iPause = EFalse; iExclusive.Signal(); iStateChange.Signal();

		iParent.PauseResumeRefresh();
		return;
	}
	
	if(iAudioThread.Create(_L("Player"), &PlayerFunc, 0x5000, iEvenBuffer.MaxLength(), 0x200000, this) == KErrNone)
	{
		iEvenBuffer.SetMax(); iOddBuffer.SetMax();
		iCurBuffer.Set(iEvenBuffer.MidTPtr(0));
		
		iAudioThread.SetPriority(EPriorityAbsoluteForegroundNormal);
		
		iAudioBusy = iSetAudioParam = EFalse;

		iLastFilePlayed = EFalse;

		iParent.SetAlive(ETrue);
		iAudioThread.Resume();
	}
	else
	{
		iAudioThread.Close();
		iAudioOutput->Stop();
		iAudioEffects->DestroyAudioEffect();
	}
}

void CPlayEngine::MaoscBufferCopied(TInt /*aError*/, const TDesC8& /*anBuffer*/)
{ iFreeBuffer.Signal(); } 

void CPlayEngine::MaoscPlayComplete(TInt aError)
{
	switch (aError)
	{
	case KErrDied:
	case KErrInUse:
	case KErrAccessDenied:
	   iParent.iAutoResumeHandler->HandlePlaybackStateChange(EFalse);
	   iParent.iAutoResumeHandler->HandlePlaybackComplete(aError);
	   Pause();
	   break;

	default:
	  iExclusive.Wait();
	  iAudioBusy = EFalse;
	  iExclusive.Signal();
	  iStateChange.Signal();
	}
}


void CPlayEngine::RunL()	// synchronously called, no synchronization needed
{
	if(iStatus != KErrNone) return;
	
	if(!iBufferPending)
	{
		if(iLastFilePlayed) iParent.SetAlive(EFalse); 

		if(iParent.iAlive)
			iParent.StartRefresh();
		else
		{
			iAudioOutput->Stop();
			iAudioEffects->DestroyAudioEffect();

			iParent.StopRefresh();
		}
		return;
	}
	
	if(iParent.iPause) iFreeBuffer.Signal();
	else
	{	
		if(iSetAudioParam)
		{
			TRAPD
			(err,
				iAudioOutput->SetDataTypeL(iAudioDataType);
				iAudioOutput->SetAudioPropertiesL(iAudioSettings.iSampleRate, iAudioSettings.iChannels)
			)

			if(err != 0)
			{
				iAudioOutput->Stop();
				iAudioEffects->DestroyAudioEffect();
				iAudioOutput->Open(&iAudioSettings);
				return;
			}

			iSetAudioParam = EFalse;
		}
		iAudioBusy = ETrue; iAudioOutput->WriteL(ActiveBuffer());
	}

	if(iParent.iNeedRefreshForPlayNextInList) iParent.BackL();

	iExclusive.Wait(); iBufferPending = EFalse; iExclusive.Signal();
	iStateChange.Signal();
}

void CPlayEngine::DoCancel()
{
}

void CPlayEngine::Start()
{
	if(iParent.iAlive) return;

	iAudioDataType = KMMFFourCCCodeNULL;
	iAudioSettings.iSampleRate = 0; iAudioSettings.iChannels = 0;
	iParent.iPause = EFalse; iAudioOutput->Open(&iAudioSettings);
}

void CPlayEngine::Stop()
{
	if(iParent.iAlive)
	{
		TRequestStatus aStatus; iAudioThread.Logon(aStatus);
		iExclusive.Wait(); iParent.SetAlive(EFalse); iExclusive.Signal();
		iFreeBuffer.Signal(10); iStateChange.Signal();	// green light!   
		User::WaitForRequest(aStatus); iAudioThread.Close();

		if(!iParent.iPause)
		{
		 iAudioOutput->Stop();
		 iAudioEffects->DestroyAudioEffect();
		}
	}
	if(IsActive()) Cancel();	// cancel an outstanding request to write to audio device
}

void CPlayEngine::Pause()
{
	if(!iParent.iPause)
	{
	iParent.iPause = ETrue;

	iAudioOutput->Stop();
	iAudioEffects->DestroyAudioEffect();

	iFreeBuffer.Signal(3);

	iParent.PauseResumeRefresh();
	}
}

void CPlayEngine::Resume()
{
	if(iParent.iPause) iAudioOutput->Open(&iAudioSettings);
}

void CPlayEngine::SetVolume(TInt aVol)
{
   if(aVol > 0)
	iAudioOutput->SetVolume(iAudioOutput->MaxVolume() * aVol / iParent.iVolumeSteps);
   else
	iAudioOutput->SetVolume(0);
}


void CPlayEngine::DolbyChange2(TBool aState)
{
	iParent.iDolbyActive = aState;
}

// End of File
