#include <e32std.h>
#include <ctsydomainpskeys.h>
#include <mpxpskeywatcher.h>

#include <AudioClientsListPSData.h>
#include <AudioClientsListPSKeys.h>

#include <voiceuidomainpskeys.h>

#include "FolderPlayView.h"
#include "autoresumehandler.h"


// CONSTANTS
const TInt KVoiceUiChangeState = 1001; // For Inform Anim Dll - dsv

const TInt KMPXErrDiedTimeout = 2000000;
// Time to wait before resume after call has ended.
const TInt KMPXResumeWaitTime = 2500000; // 2.5s


// ================= MEMBER FUNCTIONS =======================

CAutoResumeHandler::CAutoResumeHandler(CFolderView& aParent)
 :  iParent(aParent)
    {
    }


void CAutoResumeHandler::ConstructL()
    {
    // Listen to call state changes
    iStateObserver = CMPXPSKeyWatcher::NewL(KPSUidCtsyCallInformation,
                                            KCTsyCallState,this);

    // Listen to call type changes
    iTypeObserver = CMPXPSKeyWatcher::NewL(KPSUidCtsyCallInformation,
                                           KCTsyCallType,this);
    
    iVoiceCmdObserver = CMPXPSKeyWatcher::NewL( KPSUidVoiceUiAccMonitor, KVoiceUiOpenKey, this );

    iAudioPolicyObserver = CMPXPSKeyWatcher::NewL( KPSUidMMFAudioServer, KAudioPolicyAudioClients, this );

    iResumeTimer = CPeriodic::NewL(CActive::EPriorityStandard);
    }


CAutoResumeHandler* CAutoResumeHandler::NewL(CFolderView& aParent)
    {
    CAutoResumeHandler* self = new (ELeave) CAutoResumeHandler(aParent);

    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop( self );

    return self;
    }


CAutoResumeHandler::~CAutoResumeHandler()
    {
    delete iAudioPolicyObserver;
    delete iVoiceCmdObserver;
    delete iStateObserver;
    delete iTypeObserver;
    if ( iResumeTimer )
        {
        CancelResumeTimer();
        delete iResumeTimer;
        }
    }


void CAutoResumeHandler::HandleOpenFileComplete()
    {
    iPausedForCall = EFalse;
    }


void CAutoResumeHandler::HandlePlaybackStateChange(TBool aPaused)
    {
    // Any state change means that possible waiting resume is not to be done anymore.
    CancelResumeTimer();

    iPlayerOnPause = aPaused;

    if (!aPaused)
        {
        iPausedForCall = EFalse;
        }
    }


void CAutoResumeHandler::HandlePlaybackComplete(TInt aError)
    {
    iPausedForCall = EFalse;
    if ( KErrDied == aError || KErrAccessDenied == aError || KErrInUse == aError )
        {
        iKErrDiedTime.HomeTime();

        TInt callType = EPSCTsyCallTypeNone;
        TInt callState = EPSCTsyCallStateNone;
        
        if (!iTypeObserver->GetValue(callType) &&
            !iStateObserver->GetValue(callState))
            {
            if (ShouldPause())
                {
                // Getting a play complete with KErrDied here
                // means Audio Policy terminated our playback,
                // due to phone call being connected. Enable autoresume.
                iPausedForCall = ETrue;
                }
            }
        }
    }


void CAutoResumeHandler::CancelResumeTimer()
    {
    if ( iResumeTimer ) iResumeTimer->Cancel();
    }


void CAutoResumeHandler::HandlePSEvent(TUid aUid, TInt /*aKey*/)
    {
    if ( aUid == KPSUidVoiceUiAccMonitor )
        {
        DoHandleVoiceCmdChange();
        }
    else if (aUid == KPSUidMMFAudioServer)
        {
        TRAP_IGNORE(DoHandleAudioPolicyChangeL());
        }
    else
        {
        TRAP_IGNORE(DoHandleStateChangeL());
        }
    }


void CAutoResumeHandler::DoHandleStateChangeL()
    {
    TBool shouldPause = ShouldPause();
    if (shouldPause && !iPausedForCall &&
        iParent.iAlive && !iParent.iPause )
        {
        iParent.Pause();
        iPausedForCall = ETrue;
        }
    else if ( shouldPause && !iPausedForCall && iParent.iSeeking )
        {
        iParent.SeekEnd();

//        if ( iParent.iAlive && !iParent.iPause )
//            {
//            iParent.Pause();
//            iPausedForCall = ETrue;
//            }
        }
    else if(!shouldPause && iPausedForCall && iParent.iPause)
        {
        if (iResumeTimer->IsActive()) iResumeTimer->Cancel();

        iResumeTimer->Start( KMPXResumeWaitTime, KMPXResumeWaitTime,
            TCallBack(ResumeTimerCallback, this));

        iPausedForCall = EFalse;
        }
    else if ( shouldPause && iParent.iPause && !iPausedForCall &&
             iKErrDiedTime.Int64())
        {
        // Check if we recently got a playcomplete with KErrDied,
        // it was most likely caused by an active call
        TTime now;
        now.HomeTime();
        TInt64 deltaTime = now.MicroSecondsFrom(iKErrDiedTime).Int64();
        if ( deltaTime > 0 && deltaTime < KMPXErrDiedTimeout)
            {
            iResumeTimer->Cancel();
            iPausedForCall = ETrue;
            }
        }
    
    if ( shouldPause && iVoiceCmdResumeOngoing )
        {
        // Resume timer has been started after a voice command, cancel it now
        // so that playback is not resumed while calling
        if ( iResumeTimer->IsActive() )
            {
            iResumeTimer->Cancel();
            }
        iVoiceCmdResumeOngoing = EFalse;
        iPausedForCall = ETrue; // resume playback once call has been ended
        }
    }


TBool CAutoResumeHandler::ShouldPause()
    {
	TBool ret = EFalse;

        TInt callType;
        TInt callState;

        iTypeObserver->GetValue(callType);
        iStateObserver->GetValue(callState);

        if (callType == EPSCTsyCallTypeCSVoice ||
            callType == EPSCTsyCallTypeH324Multimedia ||
            callType == EPSCTsyCallTypeVoIP ||
            callType == EPSCTsyCallTypeUninitialized)
            {
            switch (callState)
                {
                case EPSCTsyCallStateAnswering:
                case EPSCTsyCallStateAlerting:
                case EPSCTsyCallStateConnected:
                case EPSCTsyCallStateDialling:
                case EPSCTsyCallStateHold:
                case EPSCTsyCallStateDisconnecting:
                    {
                    ret = ETrue;
                    break;
                    }
                case EPSCTsyCallStateRinging:
                    {
                    if (iPausedForCall)
                       {
                       ret = ETrue;
                       }
                    else
                        {
                        // Pause playback if we cannot mix music playback
                        // with ringing tone.
                        ret = ETrue;
                        }
                    break;
                    }
                default:
                    {
                    // Default is no pause
                    break;
                    }
                }
            }
    return ret;
    }


void CAutoResumeHandler::HandleResumeTimerCallback()
    {
    iVoiceCmdResumeOngoing = EFalse;
    
    CancelResumeTimer();
    TRAP_IGNORE( iParent.Resume());
    }


TInt CAutoResumeHandler::ResumeTimerCallback(TAny* aPtr)
    {
    CAutoResumeHandler* ptr = static_cast<CAutoResumeHandler*>(aPtr);

    ptr->HandleResumeTimerCallback();

    return KErrNone;
    }


void CAutoResumeHandler::DoHandleVoiceCmdChange()
{
    TInt voiceCmdState( 0 );
    iVoiceCmdObserver->GetValue( voiceCmdState );

    switch (voiceCmdState)
    {
	case KVoiceUiIsClose:
	{
		if (iPausedForVoiceCmd)
		{
			iPausedForVoiceCmd = EFalse;

			iParent.InformAniDll(KVoiceUiChangeState, EFalse);

			if ( !iPausedForCall && !iPlayerOnPause)
		        {
		            if ( iResumeTimer->IsActive() ) iResumeTimer->Cancel();
            
		            iResumeTimer->Start( KMPXResumeWaitTime, KMPXResumeWaitTime, TCallBack(ResumeTimerCallback, this) );
            
		            iPausedForVoiceCmd = EFalse;
            
		            iVoiceCmdResumeOngoing = ETrue; // flag for cancelling resume timer due to a call
		        }
		}
	} break;

	case KVoiceUiIsOpen:
	{
		if (!iPausedForVoiceCmd)
		{
			iPausedForVoiceCmd = ETrue;

			iParent.InformAniDll(KVoiceUiChangeState, ETrue);
		}
	} break;

	default:;
    }
    
    if ( iPausedForCall ) // ensure that not interfering with call handling in any circumstances
    {
	iPausedForVoiceCmd = EFalse;
    }
}


void CAutoResumeHandler::DoHandleAudioPolicyChangeL()
{
    TAudioClientList audioClients;

    TInt err(iAudioPolicyObserver->GetValue(audioClients));

    if(err == KErrNone)
    {
      if ( !iPausedForVoiceCmd && !iPausedForCall && !iPlayerOnPause)
         {
            iNumberOfActiveAudioClients = audioClients().iNumOfProcesses;                                                           

            if (!iResumeTimer->IsActive())
              iResumeTimer->Start( KMPXResumeWaitTime, KMPXResumeWaitTime, TCallBack(ResumeTimerCallback, this));
         }
    }
}

//  End of File
