/* alloc - Convenience routines for safely allocating memory
 * Copyright (C) 2007  Josh Coalson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * Ported to Symbian OS S60 3-rd ed. FP1 on May, 2009 by Alexander Demidov - demidov@bezeqint.net
 */

#ifndef FLAC__SHARE__ALLOC_H
#define FLAC__SHARE__ALLOC_H

#include <e32std.h>

#include "public/ordinals.h"

/* WATCHOUT: for c++ you may have to #define __STDC_LIMIT_MACROS 1 real early
 * before #including this file,  otherwise SIZE_MAX might not be defined
 */

#ifndef FLaC__INLINE
#define FLaC__INLINE
#endif

/* avoid User::Alloc()ing 0 bytes, see:
 * https://www.securecoding.cert.org/confluence/display/seccode/MEM04-A.+Do+not+make+assumptions+about+the+result+of+allocating+0+bytes?focusedCommentId=5407003
*/
inline void *safe_malloc_(size_t size)
{
	/* User::Alloc(0) is undefined; FLAC src convention is to always allocate */
	if(!size)
		size++;
	return User::Alloc(size);
}

inline void *safe_calloc_(size_t nmemb, size_t size)
{
	if(!nmemb || !size)
		return User::Alloc(1); // User::Alloc(0) is undefined; FLAC src convention is to always allocate
	return User::AllocZ(nmemb * size);
}

/*@@@@ there's probably a better way to prevent overflows when allocating untrusted sums but this works for now */

inline void *safe_malloc_add_2op_(size_t size1, size_t size2)
{
	size2 += size1;
	if(size2 < size1)
		return 0;
	return safe_malloc_(size2);
}


inline void *safe_malloc_mul_2op_(size_t size1, size_t size2)
{
	if(!size1 || !size2)
		return User::Alloc(1); /* User::Alloc(0) is undefined; FLAC src convention is to always allocate */
	if(size1 > SIZE_MAX / size2)
		return 0;
	return User::Alloc(size1*size2);
}


/* size1 * (size2 + size3) */
inline void *safe_malloc_muladd2_(size_t size1, size_t size2, size_t size3)
{
	if(!size1 || (!size2 && !size3))
		return User::Alloc(1); // User::Alloc(0) is undefined; FLAC src convention is to always allocate
	size2 += size3;
	if(size2 < size3)
		return 0;
	return safe_malloc_mul_2op_(size1, size2);
}

inline void *safe_realloc_mul_2op_(void *ptr, size_t size1, size_t size2)
{
	if(!size1 || !size2)
		return User::ReAlloc(ptr, 0); /* preserve POSIX User::ReAlloc(ptr, 0) semantics */
	if(size1 > SIZE_MAX / size2)
		return 0;
	return User::ReAlloc(ptr, size1*size2);
}

#endif
